@@ -1,8 +1,6 @@
-from __future__ import division, print_function, unicode_literals
-
 from ctypes import (
-    c_char_p, c_int, c_uint, c_long, c_longlong, c_size_t, c_void_p,
-    c_wchar_p, CFUNCTYPE, POINTER,
+    c_char_p, c_int, c_uint, c_long, c_longlong, c_size_t, c_int64,
+    c_void_p, c_wchar_p, CFUNCTYPE, POINTER,
 )
 
 try:
@@ -15,6 +13,7 @@ from ctypes.util import find_library
 import logging
 import mmap
 import os
+import sysconfig
 
 from .exception import ArchiveError
 
@@ -47,9 +46,14 @@ WRITE_CALLBACK = CFUNCTYPE(
 READ_CALLBACK = CFUNCTYPE(
     c_ssize_t, c_void_p, c_void_p, POINTER(c_void_p)
 )
+SEEK_CALLBACK = CFUNCTYPE(
+    c_longlong, c_int, c_void_p, c_longlong, c_int
+)
 OPEN_CALLBACK = CFUNCTYPE(c_int, c_void_p, c_void_p)
 CLOSE_CALLBACK = CFUNCTYPE(c_int, c_void_p, c_void_p)
-VOID_CB = lambda *_: ARCHIVE_OK
+
+NO_OPEN_CB = ctypes.cast(None, OPEN_CALLBACK)
+NO_CLOSE_CB = ctypes.cast(None, CLOSE_CALLBACK)
 
 
 # Type aliases, for readability
@@ -57,6 +61,11 @@ VOID_CB = lambda *_: ARCHIVE_OK
 c_archive_p = c_void_p
 c_archive_entry_p = c_void_p
 
+if sysconfig.get_config_var('SIZEOF_TIME_T') == 8:
+    c_time_t = c_int64
+else:
+    c_time_t = c_long
+
 
 # Helper functions
 
@@ -72,7 +81,7 @@ def _error_string(archive_p):
 
 def archive_error(archive_p, retcode):
     msg = _error_string(archive_p)
-    raise ArchiveError(msg, errno(archive_p), retcode, archive_p)
+    return ArchiveError(msg, errno(archive_p), retcode, archive_p)
 
 
 def check_null(ret, func, args):
@@ -101,8 +110,55 @@ def ffi(name, argtypes, restype, errcheck=None):
     return f
 
 
+def get_read_format_function(format_name):
+    function_name = 'read_support_format_' + format_name
+    func = globals().get(function_name)
+    if func:
+        return func
+    try:
+        return ffi(function_name, [c_archive_p], c_int, check_int)
+    except AttributeError:
+        raise ValueError('the read format %r is not available' % format_name)
+
+
+def get_read_filter_function(filter_name):
+    function_name = 'read_support_filter_' + filter_name
+    func = globals().get(function_name)
+    if func:
+        return func
+    try:
+        return ffi(function_name, [c_archive_p], c_int, check_int)
+    except AttributeError:
+        raise ValueError('the read filter %r is not available' % filter_name)
+
+
+def get_write_format_function(format_name):
+    function_name = 'write_set_format_' + format_name
+    func = globals().get(function_name)
+    if func:
+        return func
+    try:
+        return ffi(function_name, [c_archive_p], c_int, check_int)
+    except AttributeError:
+        raise ValueError('the write format %r is not available' % format_name)
+
+
+def get_write_filter_function(filter_name):
+    function_name = 'write_add_filter_' + filter_name
+    func = globals().get(function_name)
+    if func:
+        return func
+    try:
+        return ffi(function_name, [c_archive_p], c_int, check_int)
+    except AttributeError:
+        raise ValueError('the write filter %r is not available' % filter_name)
+
+
 # FFI declarations
 
+# library version
+version_number = ffi('version_number', [], c_int, check_int)
+
 # archive_util
 
 errno = ffi('errno', [c_archive_p], c_int)
@@ -113,10 +169,10 @@ error_string = ffi('error_string', [c_archive_p], c_char_p)
 ffi('entry_new', [], c_archive_entry_p, check_null)
 
 ffi('entry_filetype', [c_archive_entry_p], c_int)
-ffi('entry_atime', [c_archive_entry_p], c_int)
-ffi('entry_birthtime', [c_archive_entry_p], c_int)
-ffi('entry_mtime', [c_archive_entry_p], c_int)
-ffi('entry_ctime', [c_archive_entry_p], c_int)
+ffi('entry_atime', [c_archive_entry_p], c_time_t)
+ffi('entry_birthtime', [c_archive_entry_p], c_time_t)
+ffi('entry_mtime', [c_archive_entry_p], c_time_t)
+ffi('entry_ctime', [c_archive_entry_p], c_time_t)
 ffi('entry_atime_nsec', [c_archive_entry_p], c_long)
 ffi('entry_birthtime_nsec', [c_archive_entry_p], c_long)
 ffi('entry_mtime_nsec', [c_archive_entry_p], c_long)
@@ -140,10 +196,10 @@ ffi('entry_gid', [c_archive_entry_p], c_longlong)
 ffi('entry_set_size', [c_archive_entry_p, c_longlong], None)
 ffi('entry_set_filetype', [c_archive_entry_p, c_uint], None)
 ffi('entry_set_perm', [c_archive_entry_p, c_int], None)
-ffi('entry_set_atime', [c_archive_entry_p, c_int, c_long], None)
-ffi('entry_set_mtime', [c_archive_entry_p, c_int, c_long], None)
-ffi('entry_set_ctime', [c_archive_entry_p, c_int, c_long], None)
-ffi('entry_set_birthtime', [c_archive_entry_p, c_int, c_long], None)
+ffi('entry_set_atime', [c_archive_entry_p, c_time_t, c_long], None)
+ffi('entry_set_mtime', [c_archive_entry_p, c_time_t, c_long], None)
+ffi('entry_set_ctime', [c_archive_entry_p, c_time_t, c_long], None)
+ffi('entry_set_birthtime', [c_archive_entry_p, c_time_t, c_long], None)
 
 ffi('entry_update_pathname_utf8', [c_archive_entry_p, c_char_p], None)
 
@@ -160,22 +216,24 @@ READ_FORMATS = set((
 ))
 for f_name in list(READ_FORMATS):
     try:
-        ffi('read_support_format_'+f_name, [c_archive_p], c_int, check_int)
-    except AttributeError:  # pragma: no cover
-        logger.warning('read format "%s" is not supported' % f_name)
+        get_read_format_function(f_name)
+    except ValueError as e:  # pragma: no cover
+        logger.info(str(e))
         READ_FORMATS.remove(f_name)
 
 READ_FILTERS = set((
     'all', 'bzip2', 'compress', 'grzip', 'gzip', 'lrzip', 'lzip', 'lzma',
-    'lzop', 'none', 'rpm', 'uu', 'xz', 'lz4'
+    'lzop', 'none', 'rpm', 'uu', 'xz', 'lz4', 'zstd'
 ))
 for f_name in list(READ_FILTERS):
     try:
-        ffi('read_support_filter_'+f_name, [c_archive_p], c_int, check_int)
-    except AttributeError:  # pragma: no cover
-        logger.warning('read filter "%s" is not supported' % f_name)
+        get_read_filter_function(f_name)
+    except ValueError as e:  # pragma: no cover
+        logger.info(str(e))
         READ_FILTERS.remove(f_name)
 
+ffi('read_set_seek_callback', [c_archive_p, SEEK_CALLBACK], c_int, check_int)
+
 ffi('read_open',
     [c_archive_p, c_void_p, OPEN_CALLBACK, READ_CALLBACK, CLOSE_CALLBACK],
     c_int, check_int)
@@ -222,20 +280,20 @@ WRITE_FORMATS = set((
 ))
 for f_name in list(WRITE_FORMATS):
     try:
-        ffi('write_set_format_'+f_name, [c_archive_p], c_int, check_int)
-    except AttributeError:  # pragma: no cover
-        logger.warning('write format "%s" is not supported' % f_name)
+        get_write_format_function(f_name)
+    except ValueError as e:  # pragma: no cover
+        logger.info(str(e))
         WRITE_FORMATS.remove(f_name)
 
 WRITE_FILTERS = set((
     'b64encode', 'bzip2', 'compress', 'grzip', 'gzip', 'lrzip', 'lzip', 'lzma',
-    'lzop', 'uuencode', 'xz', 'lz4'
+    'lzop', 'uuencode', 'xz', 'lz4', 'zstd'
 ))
 for f_name in list(WRITE_FILTERS):
     try:
-        ffi('write_add_filter_'+f_name, [c_archive_p], c_int, check_int)
-    except AttributeError:  # pragma: no cover
-        logger.warning('write filter "%s" is not supported' % f_name)
+        get_write_filter_function(f_name)
+    except ValueError as e:  # pragma: no cover
+        logger.info(str(e))
         WRITE_FILTERS.remove(f_name)
 
 ffi('write_open',
@@ -264,5 +322,13 @@ ffi('write_fail', [c_archive_p], c_int, check_int)
 ffi('write_close', [c_archive_p], c_int, check_int)
 ffi('write_free', [c_archive_p], c_int, check_int)
 
-# library version
-ffi('version_number', [], c_int, check_int)
+# archive encryption
+
+try:
+    ffi('read_add_passphrase', [c_archive_p, c_char_p], c_int, check_int)
+    ffi('write_set_passphrase', [c_archive_p, c_char_p], c_int, check_int)
+except AttributeError:
+    logger.info(
+        f"the libarchive being used (version {version_number()}, "
+        f"path {libarchive_path}) doesn't support encryption"
+    )
